# FIXME: Streaming output is not working!
# Create customized Ollama models with modelfile (e.g.):

# # In llama3.2:1b-128k.modelfile
# FROM llama3.2:1b
# PARAMETER num_ctx 128000
# PARAMETER repeat_penalty 1.1

# Then run: ollama create llama3.2:1b-128k -f llama3.2:1b-128k.modelfile

# https://www.youtube.com/watch?v=uWDocIoiaXE

import numpy as np
from langchain_core.embeddings import Embeddings
from numpy import ndarray
from numpy.linalg import norm


def batch_generator(iterable, batch_size: int):
    """
    Splits an iterable into batches of a specified size.

    Args:
        iterable: The input iterable (e.g., generator, list).
        batch_size: The number of items per batch.

    Yields:
        Lists containing up to batch_size elements from the iterable.
    """
    batch = []
    for item in iterable:
        batch.append(item)
        if len(batch) == batch_size:
            yield batch
            batch = []
    if batch:
        yield batch


def cosine_similarities(
    embedder: Embeddings, query: str, vecs: list[np.ndarray]
) -> list[float]:
    """Computes normalized scores
    Particularly when using ChromaDB with HuggingFaceEmbeddings.
    """
    query_vec = np.array(embedder.embed_query(query))
    vecs_mat = np.stack(vecs)
    numer = vecs_mat @ query_vec
    denom: ndarray = norm(query_vec) * norm(vecs_mat, axis=1)
    return (numer / denom).tolist()



from dataclasses import dataclass
from typing import Optional

import litellm
from langchain_core.embeddings import Embeddings


@dataclass
class LiteLLMEmbeddings(Embeddings):
    model: str
    api_base: Optional[str] = None
    api_key: Optional[str] = None

    def __post_init__(self):
        self.provider, self.model_id = self.model.split("/", 1)

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embed search docs."""
        match self.provider:
            case "ollama":
                # NOTE: As of 9 May 2025, litellm.embedding somehow uses async
                # calls for ollama. This is a fix.
                from ollama import Client

                client = Client(host=self.api_base)
                return client.embed(
                    model=self.model_id,
                    input=texts,
                ).embeddings  # type: ignore

            case _:
                responses = litellm.embedding(
                    input=texts,
                    model=self.model,
                    api_base=self.api_base,
                    api_key=self.api_key,
                    aembedding=False,
                )
                return [data["embedding"] for data in responses.data]  # type: ignore

    def embed_query(self, text: str) -> list[float]:
        """Embed query text."""
        return self.embed_documents([text])[0]



# Resources:
#
# https://python.langchain.com/docs/tutorials/rag/
# https://python.langchain.com/api_reference/core/vectorstores/langchain_core.vectorstores.in_memory.InMemoryVectorStore.html
# https://python.langchain.com/docs/integrations/text_embedding/
# https://python.langchain.com/docs/how_to/chatbots_memory/

import logging
from dataclasses import dataclass
from typing import Any, Optional
from uuid import uuid4

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (
    HumanMessage,
    SystemMessage,
)
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import START, MessagesState, StateGraph

from brag.db import Database
from brag.rags.abstract import RAG_SYSTEM_PROMPT, Rag


@dataclass
class Brag(Rag):
    """Basic RAG (Brag).

    This does not rely on tool calls, so might be better for VLLM-served models
    or LLMs that aren't reliable with tool calling.
    """

    llm: BaseChatModel
    db: Database
    thread_id: Optional[str] = None
    system_prompt: str = RAG_SYSTEM_PROMPT
    verbose: bool = True

    def __post_init__(self):
        self.config = {"configurable": {"thread_id": self.thread_id or uuid4()}}
        self.apply_filter(None)

    def apply_filter(self, filter_dict: Optional[dict[str, Any]] = None):
        workflow = StateGraph(state_schema=MessagesState)

        # Define the function that calls the model
        def call_model(state: MessagesState):
            logging.debug(f"System prompt: {self.system_prompt}")
            system_message = SystemMessage(content=self.system_prompt)

            # exclude the most recent user input
            message_history = state["messages"][:-1]

            last_human_message = state["messages"][-1]
            logging.debug(f"LHM: {last_human_message.content}")

            # Summarize the messages if the chat history reaches a certain size
            if len(message_history) >= 2:
                # Invoke the model to generate conversation summary
                summary_prompt = (
                    "Given the above chat history and the latest user query below, "
                    "which might reference the chat history, "
                    "formulate a standalone question which can be understood without the "
                    "chat history. Do NOT answer the user's question. Just reformulate it "
                    "if needed; otherwise, just return it as is."
                    "\n\n# Question:\n"
                    f"{last_human_message.content}"
                )

                # Generate standalone query using chat history.
                contextualized_query = self.llm.invoke(
                    message_history + [HumanMessage(content=summary_prompt)]
                )
                self._ctx_query = contextualized_query.content

                # Retrieve context.
                self._context = self.db.retrieve(
                    str(contextualized_query.content),
                    filter_dict=filter_dict,
                )

                prompt_with_context = (
                    f"\n\n# Question:\n{contextualized_query.content}"
                    f"\n\n# Context:\n{self._context}"
                )

                # Call the model with summary & response
                response = self.llm.invoke([
                    system_message,
                    HumanMessage(content=prompt_with_context),
                ])

                # Re-add user message
                human_message = HumanMessage(content=last_human_message.content)

                updated_messages = message_history + [
                    human_message,
                    response,
                ]
            else:
                # Retrieve context.
                self._context = self.db.retrieve(
                    last_human_message.content,
                    filter_dict=filter_dict,
                )

                prompt_with_context = (
                    f"\n\n# Question:\n{last_human_message.content}"
                    f"\n\n# Context:\n{self._context}"
                )

                response = self.llm.invoke(
                    [system_message]
                    + message_history
                    + [HumanMessage(prompt_with_context)]
                )

                updated_messages = message_history + [
                    HumanMessage(last_human_message.content),
                    response,
                ]

            return {"messages": updated_messages}

        # Define the node and edge
        workflow.add_node("model", call_model)
        workflow.add_edge(START, "model")

        # Add simple in-memory checkpointer
        memory = MemorySaver()
        self.app = workflow.compile(checkpointer=memory)

    def ask(self, query: str):
        # References:
        # https://python.langchain.com/docs/how_to/qa_streaming/
        # This still only works for OpenAI but not Ollama.
        # yield "WARNING: Brag is WIP. Please use Trag for production!\n\n"
        match query:
            case "!refresh":
                self.clear_memory()
                yield "Starting a new conversation... "

            case "!cite":
                yield self._context

            case _:
                self._context = ""
                self._response = ""
                self._ctx_query = ""
                # NOTE: I don't know why, but the contextualized query is
                # repeated.
                ctx_query = {}
                for msg, _ in self.app.stream(
                    {"messages": ("user", query)},
                    stream_mode="messages",
                    config=self.config,
                ):
                    if msg.content and not isinstance(msg, HumanMessage):
                        if msg.type == "AIMessageChunk":
                            ctx_query.setdefault(msg.id, "")
                            ctx_query[msg.id] += msg.content
                            if len(ctx_query) > 1:
                                # For some reason, the LLM always responds with
                                # the contextualized question. So skip the
                                # first.
                                self._response += msg.content
                                yield msg.content

                if len(ctx_query) == 1:
                    # For some reason, the contextualized query was not included
                    # in the AI response.
                    yield ctx_query[first_key(ctx_query)]

                yield "\n\n"

                logging.info(
                    f"Contextualized Query: {self._ctx_query or query}"
                )


def first_key(d: dict[str, Any]) -> str:
    return list(d)[0]



from abc import ABC, abstractmethod
from typing import Any, Iterator, Optional
from uuid import uuid4

from langchain_core.language_models import BaseChatModel
from rich.console import Console
from rich.markdown import Markdown

from brag.db import Database
from brag.styles import make_console

"""Prompt for RAG."""
RAG_SYSTEM_PROMPT = """
Please answer the user's questions based on the provided context. If the context 
doesn't contain information relevant to the question, say you don't have enough 
information to answer the question.
""".strip().replace("\n", "")

"""Mindful prompt for RAG.

Enables the use of the LLM's existing knowledge to answer the question.
"""

MINDFUL_RAG_SYSTEM_PROMPT = """
Please answer the user's questions. A context will be provided for you 
to help you answer questions, but you need not answer based only on 
the context. If you do answer based on information that is not 
included in the context, just say so.
""".strip().replace("\n", "")

# MINDFUL_RAG_SYSTEM_PROMPT_FAILED_ATTEMP_2 = """
# Please answer the user's questions. A context will be provided for you
# to help you answer questions, but you need not answer based only on
# the context. If you still don't have enough information to answer the
# user's questions, say that you don't have enough information and if possible
# state what information you would need to answer the questions.
# """.strip().replace("\n", "")

# MINDFUL_RAG_SYSTEM_PROMPT_FAILED_ATTEMPT_1 = """
# Please answer the user's questions based on the provided context. If the
# context doesn't contain information relevant to the question, use your
# existing knowledge to answer the question. If you still don't have enough
# information to answer the question, say that you don't have enough information
# to answer the question.
# """.strip().replace("\n", "")


class Rag(ABC):  # Tool calling RAG.
    config: dict[str, Any]
    _context: str
    db: Database
    llm: BaseChatModel
    system_prompt: str
    verbose: bool = True
    console: Console = make_console()

    @abstractmethod
    def ask(self, query: str) -> Iterator[str]: ...

    @abstractmethod
    def apply_filter(self, filter_dict: Optional[dict[str, Any]] = None): ...

    def clear_memory(self) -> None:
        self.config["configurable"]["thread_id"] = uuid4()

    def print_ask(self, query: str) -> None:
        """Print query response, with context if verbose."""
        with self.console.status("[info]Generating response ..."):
            response = "".join(self.yield_ask(query))

        self.console.print(Markdown(response))

    def yield_ask(self, query: str) -> Iterator[str]:
        """Yield query response, with context if verbose."""
        yield from self.ask(query)
        if self.verbose:
            yield "\n"
            yield self._context



# https://fastapi.tiangolo.com/tutorial/static-files/
# https://www.gradio.app/guides/creating-a-custom-chatbot-with-blocks
import logging
from pathlib import Path

import gradio as gr
import uvicorn
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from pypdf import PdfReader

from brag.chat import Chatbot
from brag.rags.abstract import Rag
from brag.styles import chat_css


def make_tabs(rag: Rag, llm_name: str, corpus_dir: Path):
    with gr.Blocks(analytics_enabled=False) as rag_tab:
        with gr.Row():
            with gr.Column(scale=1):
                checkbox = gr.Checkbox(False, label="Show documents")

                def filter_docs(docnames: list[str]):
                    logging.info(f"Selected documents: {', '.join(docnames)}")
                    filter_dict = {"file_stem": {"$in": docnames}}
                    rag.apply_filter(filter_dict)

                corpus_dir = corpus_dir
                docs = {
                    doc.stem: doc
                    for doc in corpus_dir.iterdir()
                    if doc.is_file()
                }

                sorted_docnames = sorted(docs.keys())
                dropdown = gr.Dropdown(
                    choices=sorted_docnames,
                    value=sorted_docnames,
                    label="Select documents to query",
                    multiselect=True,
                    visible=False,
                )

                def toggle_dropdown(show: bool):
                    return gr.update(visible=show)

                dropdown.change(fn=filter_docs, inputs=dropdown)
                checkbox.change(
                    fn=toggle_dropdown, inputs=checkbox, outputs=dropdown
                )

            with gr.Column(scale=2, elem_classes="rag"):
                gr.Markdown(
                    f"# Chat about Documents (powered by {llm_name})",
                    elem_classes="brag-tab-header",
                )
                # FIXME: After clearing screen. Ask another question. Then the
                # old history reappears?!
                # clear = gr.HTML(
                #     "<a href='#'>Start new conversation</a>",
                #     elem_classes="brag-tab-header",
                # )

                def respond(query: str, _: list[dict[str, str]]):
                    response = ""
                    for chunk in rag.yield_ask(query):
                        response += str(chunk)
                        yield response

                chatbot = gr.Chatbot(
                    type="messages",
                    height="60vh",
                    container=False,
                    show_copy_button=True,
                )

                msg = gr.Textbox(
                    container=False,
                    placeholder="Enter query here",
                    show_copy_button=True,
                )

                gr.ChatInterface(
                    fn=respond,
                    chatbot=chatbot,
                    textbox=msg,
                    type="messages",
                    fill_height=True,
                    flagging_mode="manual",
                )

                # def clear_screen():
                #     rag.clear_memory()
                #     return "", []

                # clear.click(
                #     fn=clear_screen, inputs=None, outputs=[msg, chatbot]
                # )

        def on_load():
            # On page refresh, include all docs first.
            logging.info("Filter nothing from RAG.")
            rag.apply_filter(None)

        # If not using on_load, can comment this out.
        rag_tab.load(on_load)

    with gr.Blocks(analytics_enabled=False) as search_engine_tab:

        def clear_screen():
            return []

        gr.Markdown("# Search", elem_classes="brag-tab-header")
        clear = gr.HTML(
            "<a href='#'>Clear screen</a>", elem_classes="brag-tab-header"
        )

        chatbot = gr.Chatbot(
            type="messages",
            show_label=False,
            container=False,
            placeholder="Search history will show up here",
            height="100%",
            elem_classes="chatbot-outer",
            show_copy_button=True,
        )
        msg = gr.Textbox(
            placeholder="Enter search",
            show_label=False,
            elem_classes="user-input",
            show_copy_button=True,
        )

        def search(query: str, history: list[dict[str, str]]):
            history.append(dict(role="user", content=query))
            history.append(dict(role="assistant", content=""))

            for chunk in rag.db.retrieve(query, filter_dict=None):
                history[-1]["content"] += str(chunk)
                yield "", history

        msg.submit(search, inputs=[msg, chatbot], outputs=[msg, chatbot])
        clear.click(fn=clear_screen, inputs=None, outputs=chatbot)

    with gr.Blocks(analytics_enabled=False) as chat_tab:
        bot = Chatbot(rag.llm)

        def clear_llm_history():
            bot.clear_memory()
            gr.update(value=[])

        gr.Markdown(f"# Chat with {llm_name}", elem_classes="brag-tab-header")
        clear = gr.HTML(
            "<a href='#'>Start new conversation</a>",
            elem_classes="brag-tab-header",
        )

        chatbot = gr.Chatbot(
            type="messages",
            show_label=False,
            container=False,
            placeholder="Chat history will show up here",
            height="100%",
            elem_classes="chatbot-outer",
            show_copy_button=True,
        )

        msg = gr.Textbox(
            placeholder="Chat with BRAG ",
            show_label=False,
            elem_classes="user-input",
            show_copy_button=True,
        )

        def update(query: str, history: list[dict[str, str]]):
            history.append(dict(role="user", content=query))
            history.append(dict(role="assistant", content=""))

            for chunk in bot.stream(query):
                history[-1]["content"] += str(chunk)
                yield "", history

        msg.submit(update, inputs=[msg, chatbot], outputs=[msg, chatbot])
        clear.click(fn=clear_llm_history, inputs=None, outputs=chatbot)

    with gr.Blocks(analytics_enabled=False) as summarize_tab:
        corpus_dir = corpus_dir
        docs = {doc.stem: doc for doc in corpus_dir.iterdir() if doc.is_file()}

        def summarize(docname: str):
            prompt = "Please summarize the following:\n{content}"
            doc = docs[docname]

            logging.info("Load text.")
            match doc.suffix[1:]:
                case "pdf":
                    # content = pymupdf4llm.to_markdown(doc)
                    reader = PdfReader(doc)
                    content = "\n".join(
                        page.extract_text() for page in reader.pages
                    )
                case "txt" | "md":
                    content = doc.read_text()
                case _:
                    raise NotImplementedError()

            logging.info(
                f"Number of words in {docname}: {len(content.split())}."
            )
            response = ""
            for chunk in rag.llm.stream(prompt.format(content=content)):
                response += str(chunk.content)
                yield response

        summary = gr.Markdown(
            label="Summary",
            max_height=500,
            show_copy_button=True,
        )

        gr.Interface(
            summarize,
            [
                gr.Dropdown(
                    sorted(docs.keys()), label="Select a document to summarize"
                )
            ],
            summary,
        )

    return rag_tab, search_engine_tab, summarize_tab, chat_tab


def serve(rag: Rag, llm_name: str, corpus_dir: Path, port: int):
    # Serve /corpus dir.
    app = FastAPI()
    app.mount(
        "/corpus",
        StaticFiles(directory=str(corpus_dir)),
    )

    tabs = make_tabs(rag, llm_name=llm_name, corpus_dir=corpus_dir)
    demo = gr.TabbedInterface(
        tabs,
        ["Document Chat", "Search", "Summarize", "LLM"],
        css=chat_css,
    )
    app = gr.mount_gradio_app(app, demo, path="/")
    uvicorn.run(app, port=port)



from typing import Optional
from uuid import UUID, uuid4

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage, HumanMessage
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import START, MessagesState, StateGraph
from rich.markdown import Markdown

from brag.styles import make_console


class Chatbot:
    def __init__(self, model: BaseChatModel, thread_id: Optional[UUID] = None):
        self.model = model
        self.thread_id = thread_id or uuid4()
        self.console = make_console()

        memory = MemorySaver()

        # Define a new graph
        self.bot = (
            StateGraph(state_schema=MessagesState)
            # Define the (single) node in the graph
            .add_node("model", self.call_model)
            .add_edge(START, "model")
            # Add memory
            .compile(checkpointer=memory)
        )
        self.config = {"configurable": {"thread_id": self.thread_id}}

    def clear_memory(self):
        self.config["configurable"]["thread_id"] = uuid4()

    # Define the function that calls the model
    def call_model(self, state: MessagesState):
        response = self.model.invoke(state["messages"])
        return {"messages": response}

    def stream(self, query: str):
        if query == "!refresh":
            self.clear_memory()
            yield "Starting a new conversation!"
        else:
            input_messages = [HumanMessage(query)]

            s = self.bot.stream(
                {"messages": input_messages},
                self.config,
                stream_mode="messages",
            )
            for chunk, metadata in s:
                match chunk:
                    case AIMessage():
                        yield chunk.content

    def invoke(self, query: str, **kwargs):
        input_messages = [HumanMessage(query)]
        return self.bot.invoke(
            {"messages": input_messages}, self.config, **kwargs
        )

    def print_stream(self, query: str):
        stream = self.stream(query)

        # TODO: deprecate
        # for chunk in stream:
        #     print(chunk, end="")

        with self.console.status("[info]Generating response ..."):
            response = "".join(stream)
            self.console.print(Markdown(response))

        print()



# TODO: Deprecate?

from dataclasses import dataclass
from functools import cached_property
from math import ceil

_MAX_CHUNKS_SENT = 30000
_MAX_CTX_CHUNKS = 5500


@dataclass
class Defaults:
    chunk_size: int = 512
    temperature: float = 0.2
    max_chunks_sent: int = _MAX_CHUNKS_SENT
    max_ctx_chunks: int = _MAX_CTX_CHUNKS

    @staticmethod
    def chunk_overlap_from(chunk_size: int) -> int:
        return ceil(chunk_size * 0.2)

    @staticmethod
    def batch_size_from(
        chunk_size: int, max_chunks_sent: int = _MAX_CHUNKS_SENT
    ) -> int:
        return max_chunks_sent // chunk_size

    @staticmethod
    def num_retrieved_docs_from(
        chunk_size: int, max_ctx_chunks: int = _MAX_CTX_CHUNKS
    ) -> int:
        return max_ctx_chunks // chunk_size

    @cached_property
    def chunk_overlap(self) -> int:
        return self.chunk_overlap_from(self.chunk_size)

    @cached_property
    def batch_size(self) -> int:
        return self.max_chunks_sent // self.chunk_size

    @cached_property
    def num_retrieved_docs(self) -> int:
        return self.max_ctx_chunks // self.chunk_size



# Experimental WIP.
# https://python.langchain.com/docs/how_to/indexing/

from langchain.indexes import SQLRecordManager, index
from langchain_core.documents import Document


class Index:
    def __init__(
        self,
        vectorstore,
        db: str = "chromadb",
        source_id_key: str = "source",
    ):
        self.vectorstore = vectorstore
        self.collection_name = self.vectorstore._collection.name
        self.db = db
        self.namespace = f"{self.db}/{self.collection_name}"
        self.record_manager = SQLRecordManager(
            self.namespace, db_url="sqlite:///record_manager_cache.sql"
        )
        self.source_id_key = source_id_key
        self.record_manager.create_schema()
        self._index = index

    def count(self) -> int:
        return self.vectorstore._collection.count()

    def add(self, docs: list[Document]):
        return self._index(
            docs,
            self.record_manager,
            self.vectorstore,
            cleanup="incremental",
            source_id_key=self.source_id_key,
        )

    def sync(self, docs: list[Document]):
        return self._index(
            docs,
            self.record_manager,
            self.vectorstore,
            cleanup="full",
            source_id_key=self.source_id_key,
        )

    def clear(self):
        """Helper method to clear content. See the `full` mode section to
        to understand why it works."""
        return self.sync([])



# Resources:
#
# https://python.langchain.com/docs/tutorials/rag/
# https://python.langchain.com/api_reference/core/vectorstores/langchain_core.vectorstores.in_memory.InMemoryVectorStore.html
# https://python.langchain.com/docs/integrations/text_embedding/
#
# https://python.langchain.com/docs/how_to/self_query/

import logging
from dataclasses import dataclass
from typing import Any, Iterator, Optional
from uuid import uuid4

from langchain_core.language_models import BaseChatModel
from langchain_core.messages import HumanMessage
from langchain_core.tools import tool
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import create_react_agent

from brag.db import Database
from brag.rags.abstract import RAG_SYSTEM_PROMPT, Rag


@dataclass
class Trag(Rag):  # Tool calling RAG.
    llm: BaseChatModel
    db: Database
    thread_id: Optional[str] = None
    system_prompt: str = RAG_SYSTEM_PROMPT
    verbose: bool = True

    def apply_filter(self, filter_dict: Optional[dict[str, Any]] = None):
        self.make_retrieve_tool(filter_dict)  # Make document retriever tool.
        self.make_executor()  # Make langchain agent executor.

    def __post_init__(self):
        self.config = {"configurable": {"thread_id": self.thread_id or uuid4()}}
        self.apply_filter(None)

    def make_retrieve_tool(self, filter_dict: Optional[dict[str, Any]]):
        # NOTE: A docstring is required here for tool calling!
        def retrieve(query: str):
            """Retrieve information related to a query."""
            return self.db.retrieve(query, filter_dict=filter_dict)

        self.retrieve_tool = tool(response_format="content")(retrieve)

    def make_executor(self):
        memory = MemorySaver()
        tools = [self.retrieve_tool]
        logging.debug(f"System prompt: {self.system_prompt}")
        self.agent_executor = create_react_agent(
            self.llm,
            tools=tools,
            checkpointer=memory,
            prompt=self.system_prompt,
        )

    def ask(self, query: str) -> Iterator[str]:
        # References:
        # https://python.langchain.com/docs/how_to/qa_streaming/
        # This still only works for OpenAI but not Ollama.
        match query:
            case "!refresh":
                self.clear_memory()
                yield "Starting a new conversation... "

            case "!cite":
                yield self._context

            case _:
                self._context = ""
                self._response = ""
                for msg, _ in self.agent_executor.stream(
                    {"messages": ("user", query)},
                    stream_mode="messages",
                    config=self.config,
                ):
                    if msg.content and not isinstance(msg, HumanMessage):
                        if msg.type == "tool":
                            self._context = msg.content
                        else:
                            self._response += msg.content
                            # print(msg.content, flush=True, end="|")
                            yield msg.content



import logging
import sqlite3
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional

from chromadb import EphemeralClient, HttpClient, PersistentClient

# from langchain.chains.query_constructor.schema import AttributeInfo
from langchain.text_splitter import (
    MarkdownTextSplitter,
    RecursiveCharacterTextSplitter,
)
from langchain_chroma import Chroma
from langchain_community.document_loaders import (
    Docx2txtLoader,
    PyMuPDFLoader,
    TextLoader,
    UnstructuredXMLLoader,
)
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from rich.progress import track

from brag.defaults import Defaults
from brag.util import batch_generator


@dataclass
class MtimeDB:
    """An Sqlite3 db interface for last corpus doc modification times."""

    dbpath: Path
    table_name: str

    def open(self):
        self.dbpath.parent.mkdir(parents=True, exist_ok=True)
        self.con = sqlite3.connect(self.dbpath, autocommit=True)
        self.cur = self.con.cursor()
        self.cur.execute(
            f"CREATE TABLE IF NOT EXISTS {self.table_name}(filename, stem, mtime)"
        )

    def close(self):
        self.con.close()

    def drop(self):
        self.cur.execute(f"DROP TABLE IF EXISTS {self.table_name}")

    def create(self):
        self.open()
        self.close()

    def insert(self, file: Path):
        self.open()
        self.cur.execute(
            f"INSERT INTO {self.table_name} VALUES(?, ?, ?)",
            (file.name, file.stem, file.stat().st_mtime),
        )
        self.close()

    def update(self, file: Path):
        self.open()
        self.cur.execute(
            f"Update {self.table_name} SET mtime = ? WHERE filename = ?",
            (file.name, file.stat().st_mtime),
        )
        self.close()

    def insertmany(self, files: Iterable[Path]):
        data = ((file.name, file.stem, file.stat().st_mtime) for file in files)
        self.open()
        self.cur.executemany(
            f"INSERT INTO {self.table_name} VALUES(?, ?, ?)", data
        )
        self.close()

    def get_mtime(self, file: Path):
        """Return the modification time of the file when indexed.

        If not indexed, will return -1.
        """
        self.open()
        record = self.cur.execute(
            f"SELECT * FROM {self.table_name} WHERE filename='{file.name}'"
        ).fetchone()
        self.close()

        return -1 if record is None else record[2]

    def get_filenames_in_index(self) -> list[str]:
        self.open()
        records = self.cur.execute(
            f"SELECT filename FROM {self.table_name}"
        ).fetchall()
        self.close()
        return [record[0] for record in records]


def get_page_or_chunk(doc: Document):
    m = doc.metadata
    return f"p. {m['page']}" if "page" in m else f"chunk {m['chunk']}"


class Database:
    def __init__(
        self,
        corpus_dir: Path,
        embedder: Embeddings,
        chunk_size: Optional[int] = None,
        chunk_overlap: Optional[int] = None,
        index_doc_batch_size: Optional[int] = None,
        num_retrieved_docs: Optional[int] = None,
        cache_db: bool = True,
        db_dir: Optional[str] = None,
        db_name: Optional[str] = None,
        db_host: Optional[str] = None,
        db_port: Optional[int] = None,
        mtime_db_path: Optional[str] = None,
        terminal: bool = False,
        mtime_table_name: str = "corpus_mtimes",
        min_space_to_bang: int = 1,
    ):
        self.corpus_dir = corpus_dir
        self.embedder = embedder
        self.cache_db = cache_db

        defaults = Defaults()
        self.chunk_size = chunk_size or defaults.chunk_size
        self.chunk_overlap = chunk_overlap or defaults.chunk_overlap_from(
            self.chunk_size
        )
        self.index_doc_batch_size = (
            index_doc_batch_size or defaults.batch_size_from(self.chunk_size)
        )
        self.num_retrieved_docs = (
            num_retrieved_docs
            or defaults.num_retrieved_docs_from(self.chunk_size)
        )

        self.db_name = db_name or "brag-db"
        self.db_dir = Path(db_dir or ".brag/db")
        self.db_host = db_host
        self.db_port = db_port

        self.mtime_db_path = mtime_db_path
        self.terminal = terminal
        self.mtime_table_name = mtime_table_name
        self.mtime_db = MtimeDB(
            dbpath=self.mtime_db_path_, table_name=self.mtime_table_name
        )

        # Minimum allowed ratio of space to exclaimation points.  If
        # min_space_to_bang is 10, that means I expect/require 10 ' ' for every
        # '!'. This can be violated if the document is corrupt.
        self.min_space_to_bang = min_space_to_bang

        logging.info(f"chunksize: {self.chunk_size}")
        logging.info(f"chunkoverlap: {self.chunk_overlap}")
        self.index()  # Index corpus.

    @property
    def in_sync(self) -> bool:
        return self.corpus_dir_mtime_when_last_indexed == self.corpus_dir_mtime

    @property
    def index_info_path(self):
        return self.db_dir / "index-info.txt"

    @property
    def corpus_dir_mtime_when_last_indexed(self) -> float:
        # Returns: when the corpus was last indexed, what was the mtime of the
        # corpus directory?
        self.index_info_path.parent.mkdir(parents=True, exist_ok=True)
        if self.index_info_path.exists():
            return float(self.index_info_path.read_text())
        else:
            # Record.
            self.index_info_path.write_text(str("-1"))
            return -1

    @property
    def corpus_dir_mtime(self) -> float:
        return self.corpus_dir.stat().st_mtime

    @property
    def mtime_db_path_(self) -> Path:
        """Sqlite3 db containing last modification times of corpus docs."""
        return Path(
            self.mtime_db_path
            or f"{self.db_dir}/{self.mtime_table_name}.sqlite3"
        )

    @property
    def bar_format(self):
        return "{desc} [{percentage:3.0f}% | {n_fmt}/{total_fmt} | {elapsed}<{remaining} | {rate_fmt}]"

    def delete_collection(self):
        match self.vectorstore:
            case Chroma():
                self.vectorstore.delete_collection()
                self.mtime_db.drop()

    def index(self):
        """Sync vectorstore and docs in corpus.

        - If a file is in corpus but not in vectorstore, the file is indexed.
        - If a file was in indexed but no longer in corpus, the file is removed
          from index.
        - If file is in corpus and has been indexed, but the file has been
          modified since indexing, the file is reindexed (i.e, removed from
          index, then indexed).
        """
        if not self.in_sync:  # no need to index if in sync.
            # Record the current mtime of the corpus directory.
            self.index_info_path.write_text(str(self.corpus_dir_mtime))

            # Index everything, if needed. docs will contain only documents that
            # are out of sync. So, documents that are in sync will not be
            # reindexed. New files in corpus will also be indexed.
            docs = self.parse_docs()
            batches = list(batch_generator(docs, self.index_doc_batch_size))
            for batch in track(batches, description="Indexing"):
                self.vectorstore.add_documents(batch)

            # Remove documents that are in index but no longer in corpus dir.
            fnames_index = set(self.mtime_db.get_filenames_in_index())
            fnames_corpus = {file.name for file in self.corpus_dir.iterdir()}
            files_to_remove = fnames_index.difference(fnames_corpus)
            for fname in files_to_remove:
                logging.info(f"Removing {fname} from index.")
                self.vectorstore.delete(where=dict(filename=fname))

    @cached_property
    def vectorstore(self):
        if self.cache_db:
            self.db_dir.mkdir(parents=True, exist_ok=True)

            if self.db_port is None:
                client = PersistentClient(path=str(self.db_dir))
            else:
                client = HttpClient(
                    host=self.db_host or "localhost", port=self.db_port
                )

        else:
            client = EphemeralClient()

        return Chroma(
            collection_name=self.db_name,
            embedding_function=self.embedder,
            client=client,
            collection_metadata={"hnsw:space": "cosine"},
        )

    @cached_property
    def vectorstore_isempty(self) -> bool:
        match self.vectorstore:
            case Chroma():
                return self.vectorstore._collection.count() == 0
            case _:
                raise NotImplementedError()

    def parse_file(self, file: Path) -> Iterator[Document]:
        mtime_when_indexed = self.mtime_db.get_mtime(file)
        logging.debug(f"{file.stem}, {mtime_when_indexed}")
        if file.stat().st_mtime > mtime_when_indexed:
            # need to index / reindex.
            if mtime_when_indexed < 0:
                self.mtime_db.insert(file)
            else:
                self.mtime_db.update(file)

            try:
                match file.suffix.lower():  # file extension.
                    case ".pdf":
                        loader = PyMuPDFLoader(file)
                        text_splitter = RecursiveCharacterTextSplitter(
                            chunk_size=self.chunk_size,
                            chunk_overlap=self.chunk_overlap,
                        )

                    case ".md" | ".markdown":
                        loader = TextLoader(file)
                        text_splitter = MarkdownTextSplitter(
                            chunk_size=self.chunk_size,
                            chunk_overlap=self.chunk_overlap,
                        )

                    case ".docx":
                        loader = Docx2txtLoader(file)
                        text_splitter = RecursiveCharacterTextSplitter(
                            chunk_size=self.chunk_size,
                            chunk_overlap=self.chunk_overlap,
                        )

                    case ".xml" | ".nxml":
                        loader = UnstructuredXMLLoader(file)
                        text_splitter = RecursiveCharacterTextSplitter(
                            chunk_size=self.chunk_size,
                            chunk_overlap=self.chunk_overlap,
                        )

                    case _:
                        # Assume text file.
                        loader = TextLoader(file)
                        text_splitter = RecursiveCharacterTextSplitter(
                            chunk_size=self.chunk_size,
                            chunk_overlap=self.chunk_overlap,
                        )

                docs = loader.load_and_split(text_splitter=text_splitter)

                # Add additional metadata.
                for i, d in enumerate(docs):
                    bang_count = d.page_content.count("!")
                    space_count = d.page_content.count(" ")
                    space_to_bang = space_count / (bang_count + 1e-6)

                    # Some documents are '!' delimited instead of ' '
                    # delimited. This messes with some vector embedders.  The
                    # following attempts to infer whether such a case has
                    # occured by computing the ratio of ' ' to '!'. By default,
                    # if '!' absolutely shows up more frequently than ' ', then
                    # all ' ' will be replaced by '!'. min_space_to_bang can be
                    # user specified.  If not specified, will default to 1,
                    # which implies a 1:1 ratio is acceptable (but 1:2, 1:3, etc
                    # is not).
                    if (
                        bang_count > 1
                        and space_to_bang < self.min_space_to_bang
                    ):
                        print(
                            "brag detected too many '!'. "
                            f"{file.name} Could be a problematic document."
                            "Converting '!' to ' '."
                            "If this is expected, try decreasing `min_space_to_bang` "
                            "in brag.db.Database."
                        )
                        print("space_to_bang:", space_to_bang)
                        print(d.page_content)
                        d.page_content = d.page_content.replace("!", " ")

                    d.metadata["chunk"] = i + 1
                    d.metadata["file_name"] = file.name
                    d.metadata["file_stem"] = file.stem
                    d.metadata["abs_path"] = str(file.absolute())
                    d.metadata["corpus_dir"] = str(self.corpus_dir)
                    d.metadata["rel_path"] = str(
                        file.relative_to(self.corpus_dir)
                    )
                    d.metadata["db_name"] = self.db_name

                yield from docs

            except Exception:
                logging.warning(f"Skipping {file.name} due to parsing errors.")

    def delete_file(self, file_name: str):
        logging.info(f"deleting file: {file_name}")
        self.vectorstore.delete(where={"file_name": file_name})
        logging.info("done.")

    def add_file(self, file_name: str):
        logging.info("adding file: ", file_name)
        file = self.corpus_dir / file_name
        docs = self.parse_file(file)
        batches = list(batch_generator(docs, self.index_doc_batch_size))
        for batch in track(batches, description="Indexing"):
            self.vectorstore.add_documents(batch)

        logging.info("done.")

    def parse_docs(self) -> Iterator[Document]:
        # TODO: Provide an option to parallelize this.
        num_files = self.count_files_in_dir(self.corpus_dir)
        for file in track(
            self.corpus_dir.iterdir(),
            total=num_files,
            description="Parsing ",
        ):
            if file.is_file():
                yield from self.parse_file(file)

    def count_files_in_dir(self, dir: Path):
        return sum(1 for f in dir.iterdir() if f.is_file())

    def retrieve(self, query: str, filter_dict: Optional[dict[str, Any]]):
        # Scores are in [0, 1]. 0 for most dissimilar. 1 for most similar.
        retrieved_docs, scores = zip(
            *(
                self.vectorstore.similarity_search_with_relevance_scores(
                    query, k=self.num_retrieved_docs, filter=filter_dict
                )
            )
        )
        serialized_context = serialize(retrieved_docs, scores, self.terminal)  # type: ignore
        return serialized_context

    # def sync(self): ...


def serialize(docs: list[Document], scores: list[float], terminal: bool):
    num_sources = len(docs)

    if terminal:
        serialized = "\n\n".join(
            (
                f"{i + 1}. {doc.metadata['file_name']} ({get_page_or_chunk(doc)})"
                f"\nExcerpt ({score * 100:3.0f}%): {doc.page_content}"
            )
            for i, (doc, score) in enumerate(zip(docs, scores))
        )
        return f"\nSources ({num_sources})\n\n***\n\n" + serialized
    else:
        tag_open = (
            f"<details><summary><b>Sources ({num_sources})</b></summary>"
            "<div class='sources'>"
            "<ol class='brag-source'>"
        )
        body = ""
        tag_close = "</ol></div></details>"

        for doc, score in zip(docs, scores):
            m = doc.metadata
            if "page" in m:
                page = m["page"] + 1
                location = f"p. {page}"
                rel_path = m["rel_path"] + f"#page={page}"
            else:
                location = f"chunk {m['chunk']}"
                rel_path = m["rel_path"]

            fname = m["file_name"]
            fstem = Path(fname).stem
            url = f"/corpus/{rel_path}"

            body += (
                "<li>"
                f"<a href='{url}' target=_blank>{fstem} ({location})</a>"
                "<details>"
                f"<summary>Preview (Score: {score * 100:.0f}%)</summary>"
                f"<div class='preview'>{doc.page_content}</div>"
                "</details>"
                "</li>"
            )

        return tag_open + body + tag_close



"""Basic RAG (brag)."""

# NOTE: Putting the import statements near usage decreases time taken to
# generate help menus.

import logging
from importlib.metadata import version
from pathlib import Path
from textwrap import dedent
from typing import Annotated, Literal, Optional

from typer import Option, Typer

__version__ = version("pybrag")


def set_logging(log_level: str):
    numeric_level = getattr(logging, log_level.upper(), None)
    logging.basicConfig(
        level=numeric_level, format="[BRAG-%(levelname)s] %(message)s"
    )


app = Typer()


@app.command()
def index(
    docs_dir: Annotated[
        Path,
        Option(
            "-d",
            "--dir",
            help="Directory containing documents. Must be a flat directory. Hidden files will not be indexed.",
        ),
    ] = Path("."),
    emb: Annotated[
        str,
        Option(
            help=dedent(
                """\
                Embedding model in litellm format (provider/model_id). e.g.,
                * openai/text-embedding-3-small
                * ollama/nomic-embed-text
                * hosted_vllm/BAAI/bge-m3
                """
            )
        ),
    ] = "openai/text-embedding-3-small",
    emb_base_url: Annotated[
        Optional[str],
        Option(
            help=dedent(
                """\
                Base url for embedder. For example, 
                * https://api.openai.com/v1 (openai)
                * http://localhost:11434/v1 (ollama)
                """
            )
        ),
    ] = None,
    emb_api_key: Annotated[
        Optional[str],
        Option(
            envvar=["BRAG_EMB_API_KEY", "OPENAI_API_KEY"],
            help="API key for embedder",
        ),
    ] = None,
    chunk_size: Annotated[
        int, Option(help="Document chunksize for embedding documents")
    ] = 1000,
    chunk_overlap: Annotated[
        int, Option(help="Document chunk overlap for embedding documents")
    ] = 100,
    batchsize: Annotated[int, Option(help="Batchsize for embedding")] = 10,
    ssl_verify: Annotated[
        bool, Option(help="Whether or not to enable ssl checks")
    ] = True,
    db_dir: Annotated[
        Path, Option("--db", help="Directory to store vector database")
    ] = Path(".brag") / "db",
    db_name: Annotated[
        str, Option(help="Database collection name")
    ] = "brag-db",
    db_host: Annotated[
        Optional[str],
        Option(
            help=dedent("""\
                Hostname for chromadb server. If not provided and
                --db-port is provided, this will default to 'localhost'.
                If --db-port is not provided, this flag is ignored. 
                """)
        ),
    ] = None,
    db_port: Annotated[
        Optional[int],
        Option(
            help=dedent("""\
                Port for chromadb server. If not supplied, chromadb will
                run locally.
                """)
        ),
    ] = None,
    log_level: str = "WARNING",
    min_space_to_bang: Annotated[
        int,
        Option(
            help=dedent(
                """\
                Some documents are '!' delimited instead of ' ' delimited. This
                messes with some vector embedders.  This option attempts to
                infer whether such a case has occured by computing the ratio of
                ' ' to '!'. By default, if '!' absolutely shows up more
                frequently than ' ', then all ' ' will be replaced by '!'.
                min_space_to_bang can be user specified.  If not specified, will
                default to 1, which implies a 1:1 ratio is acceptable (but 1:2,
                1:3, etc is not).
                """
            )
        ),
    ] = 1,
):
    import litellm

    from brag.db import Database
    from brag.emb import LiteLLMEmbeddings

    if not ssl_verify:
        litellm.ssl_verify = False

    set_logging(log_level)

    db_dir.mkdir(parents=True, exist_ok=True)

    embedder = LiteLLMEmbeddings(
        model=emb,
        api_base=emb_base_url,
        api_key=emb_api_key,
    )
    Database(
        corpus_dir=Path(docs_dir),
        embedder=embedder,
        index_doc_batch_size=batchsize,
        chunk_overlap=chunk_overlap,
        chunk_size=chunk_size,
        cache_db=True,
        db_dir=str(db_dir),
        db_name=db_name,
        db_host=db_host,
        db_port=db_port,
        min_space_to_bang=min_space_to_bang,
    )


@app.command()
def ask(
    docs_dir: Annotated[
        Path,
        Option(
            "-d",
            "--dir",
            help=(
                "Directory containing documents. Must be a flat directory. "
                "Hidden files will not be indexed."
            ),
        ),
    ] = Path("."),
    llm: Annotated[
        str,
        Option(
            help=dedent(
                """\
                LLM in litellm format (provider/model_id). e.g.,
                * openai/o3
                * ollama/gpt-oss:120b
                * hosted_vllm/meta-llama/Llama-3.1-8B-Instruct
                """
            )
        ),
    ] = "openai/o3",
    llm_base_url: Annotated[
        Optional[str],
        Option(
            help=dedent(
                """\
                Base url for embedder. For example, 
                * https://api.openai.com/v1 (openai)
                * http://localhost:11434/v1 (ollama)
                """
            )
        ),
    ] = None,
    llm_api_key: Annotated[
        Optional[str],
        Option(
            envvar=["BRAG_LLM_API_KEY", "OPENAI_API_KEY"],
            help="API key for embedder",
        ),
    ] = None,
    temperature: Annotated[
        Optional[float], Option(help="Temperature for LLM.", max=1.0, min=0.0)
    ] = None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    emb: Annotated[
        str,
        Option(
            help=dedent(
                """\
                Embedding model in litellm format (provider/model_id). e.g.,
                * openai/text-embedding-3-small
                * ollama/nomic-embed-text
                * hosted_vllm/BAAI/bge-m3
                """
            )
        ),
    ] = "openai/text-embedding-3-small",
    emb_base_url: Annotated[
        Optional[str],
        Option(
            help=dedent(
                """\
                Base url for embedder. For example, 
                * https://api.openai.com/v1 (openai)
                * http://localhost:11434/v1 (ollama)
                """
            )
        ),
    ] = None,
    emb_api_key: Annotated[
        Optional[str],
        Option(
            envvar=["BRAG_EMB_API_KEY", "OPENAI_API_KEY"],
            help="API key for embedder",
        ),
    ] = None,
    chunk_size: Annotated[
        int, Option(help="Document chunksize for embedding documents")
    ] = 1000,
    chunk_overlap: Annotated[
        int, Option(help="Document chunk overlap for embedding documents")
    ] = 100,
    batchsize: Annotated[int, Option(help="Batchsize for embedding")] = 10,
    ssl_verify: Annotated[
        bool, Option(help="Whether or not to enable ssl checks")
    ] = True,
    db_dir: Annotated[
        Path, Option("--db", help="Directory to store vector database")
    ] = Path(".brag") / "db",
    db_name: Annotated[
        str, Option(help="Database collection name")
    ] = "brag-db",
    db_host: Annotated[
        Optional[str],
        Option(
            help=dedent("""\
                Hostname for chromadb server. If not provided and
                --db-port is provided, this will default to 'localhost'.
                If --db-port is not provided, this flag is ignored. 
                """)
        ),
    ] = None,
    db_port: Annotated[
        Optional[int],
        Option(
            help=dedent("""\
                Port for chromadb server. If not supplied, chromadb will
                run locally.
                """)
        ),
    ] = None,
    cache_db: Annotated[
        bool,
        Option(
            help="Whether or not to store vector database to disk",
        ),
    ] = True,
    num_retrieved_docs: Annotated[
        Optional[int],
        Option(
            help=dedent(
                """\
                number of document chunks to retrieve. Recommendation: Use an
                LLM with a context window of at least 65K (e.g. the llama3.1,
                llama3.2, llama3.3.).  Number of retrieved documents should
                result in approx 5000-6000 chunks.  e.g. if using a chunksize of
                340, using about 15 retrieved docs would result in ~5100 chunks.
                batchsize * chunksize should be about 50000 chunks.  So if using
                a chunksize of 340, use a batch size of about 147.
                """
            ),
        ),
    ] = 10,
    rag_type: Annotated[
        Literal["brag", "trag"],
        Option(
            help="Port to serve web app. If not supplied, runs only in terminal."
        ),
    ] = "brag",
    port: Annotated[
        Optional[int],
        Option(
            help="Port to serve web app. If not supplied, runs only in terminal."
        ),
    ] = None,
    system_prompt_path: Annotated[
        Optional[Path],
        Option(
            help=dedent("""\
            Optional path to system prompt. The system prompt should be in a
            text file. If supplied, the --system-prompt flag will have no
            effect. If not supplied, then --system-prompt is applied.

            Example usage:
                brag ask --system-prompt-path=prompt.txt
            """),
        ),
    ] = None,
    system_prompt: Annotated[
        Literal["basic", "mindful"],
        Option(
            help=dedent("""\
            System prompt to use. Overriden if --system-prompt-path is supplied.
              * basic: answers based only on context.
              * mindful: answers based on context and llm's knowledge
            """)
        ),
    ] = "basic",
    verbose: Annotated[
        bool,
        Option(
            help=dedent(
                """\
            Whether or not to show rag sources. If present, sources will be
            shown in command line. Sources can still be recovered via !cite.
            However, sources are always presented in the web app.
            """
            ),
        ),
    ] = False,
    log_level: str = "WARNING",
    min_space_to_bang: Annotated[
        int,
        Option(
            help=dedent(
                """\
                Some documents are '!' delimited instead of ' ' delimited. This
                messes with some vector embedders.  This option attempts to
                infer whether such a case has occured by computing the ratio of
                ' ' to '!'. By default, if '!' absolutely shows up more
                frequently than ' ', then all ' ' will be replaced by '!'.
                min_space_to_bang can be user specified.  If not specified, will
                default to 1, which implies a 1:1 ratio is acceptable (but 1:2,
                1:3, etc is not).
                """
            )
        ),
    ] = 1,
):
    import litellm
    from langchain_litellm import ChatLiteLLM

    from brag.db import Database
    from brag.emb import LiteLLMEmbeddings

    if not ssl_verify:
        litellm.ssl_verify = False

    set_logging(log_level)

    chat_model = ChatLiteLLM(
        model=llm,
        api_key=llm_api_key,
        api_base=llm_base_url,
        temperature=temperature,
        max_completion_tokens=max_completion_tokens,
        max_tokens=max_tokens,
    )
    logging.info(f"Embedder: {emb}, base url: {emb_base_url}")

    embedder = LiteLLMEmbeddings(
        model=emb,
        api_base=emb_base_url,
        api_key=emb_api_key,
    )

    db = Database(
        corpus_dir=Path(docs_dir),
        embedder=embedder,
        index_doc_batch_size=batchsize,
        num_retrieved_docs=num_retrieved_docs,
        chunk_overlap=chunk_overlap,
        chunk_size=chunk_size,
        cache_db=cache_db,
        db_dir=str(db_dir),
        db_name=db_name,
        db_host=db_host,
        db_port=db_port,
        # Always show sources for web app.
        terminal=port is None,
        min_space_to_bang=min_space_to_bang,
    )
    match rag_type:
        case "brag":
            from brag.rags import Brag

            ChosenRag = Brag
        case "trag":
            from brag.rags import Trag

            ChosenRag = Trag

    match system_prompt_path, system_prompt:
        case None, "basic":
            from brag.rags.abstract import RAG_SYSTEM_PROMPT

            _system_prompt = RAG_SYSTEM_PROMPT
        case None, "mindful":
            from brag.rags.abstract import MINDFUL_RAG_SYSTEM_PROMPT

            _system_prompt = MINDFUL_RAG_SYSTEM_PROMPT
        case Path(), _:  # system_prompt_path is supplied.
            logging.debug(
                f"system_prompt_path is supplied and is:{system_prompt_path}"
            )
            _system_prompt = system_prompt_path.read_text()
            logging.debug(_system_prompt)

    rag = ChosenRag(
        chat_model,
        db,
        thread_id=None,
        system_prompt=_system_prompt,
        # Always show sources for web app.
        verbose=verbose or port is not None,
    )
    if port:
        from brag.ui import serve

        serve(rag, llm_name=llm, corpus_dir=docs_dir, port=port)
    else:
        from brag.repl import AskREPL

        AskREPL().run(
            "Ask questions about your corpus!",
            rag.print_ask,
        )


@app.command()
def search(
    docs_dir: Annotated[
        Path,
        Option(
            "-d",
            "--dir",
            help="Directory containing documents. Must be a flat directory. Hidden files will not be indexed.",
        ),
    ] = Path("."),
    emb: Annotated[
        str,
        Option(
            help=dedent(
                """\
                Embedding model in litellm format (provider/model_id). e.g.,
                * openai/text-embedding-3-small
                * ollama/nomic-embed-text
                * hosted_vllm/BAAI/bge-m3
                """
            )
        ),
    ] = "openai/text-embedding-3-small",
    emb_base_url: Annotated[
        Optional[str],
        Option(
            help=dedent(
                """\
                Base url for embedder. For example, 
                * https://api.openai.com/v1 (openai)
                * http://localhost:11434/v1 (ollama)
                """
            )
        ),
    ] = None,
    emb_api_key: Annotated[
        Optional[str],
        Option(
            envvar=["BRAG_EMB_API_KEY", "OPENAI_API_KEY"],
            help="API key for embedder",
        ),
    ] = None,
    chunk_size: Annotated[
        int, Option(help="Document chunksize for embedding documents")
    ] = 1000,
    chunk_overlap: Annotated[
        int, Option(help="Document chunk overlap for embedding documents")
    ] = 100,
    batchsize: Annotated[int, Option(help="Batchsize for embedding")] = 10,
    ssl_verify: Annotated[
        bool, Option(help="Whether or not to enable ssl checks")
    ] = True,
    db_dir: Annotated[
        Path, Option("--db", help="Directory to store vector database")
    ] = Path(".brag") / "db",
    db_name: Annotated[
        str, Option(help="Database collection name")
    ] = "brag-db",
    db_host: Annotated[
        Optional[str],
        Option(
            help=dedent("""\
                Hostname for chromadb server. If not provided and
                --db-port is provided, this will default to 'localhost'.
                If --db-port is not provided, this flag is ignored. 
                """)
        ),
    ] = None,
    db_port: Annotated[
        Optional[int],
        Option(
            help=dedent("""\
                Port for chromadb server. If not supplied, chromadb will
                run locally.
                """)
        ),
    ] = None,
    cache_db: Annotated[
        bool,
        Option(
            help="Whether or not to store vector database to disk",
        ),
    ] = True,
    num_retrieved_docs: Annotated[
        Optional[int],
        Option(
            help=dedent(
                """\
                number of document chunks to retrieve. Recommendation: Use an
                LLM with a context window of at least 65K (e.g. the llama3.1,
                llama3.2, llama3.3.).  Number of retrieved documents should
                result in approx 5000-6000 chunks.  e.g. if using a chunksize of
                340, using about 15 retrieved docs would result in ~5100 chunks.
                batchsize * chunksize should be about 50000 chunks.  So if using
                a chunksize of 340, use a batch size of about 147.
                """
            ),
        ),
    ] = 10,
    log_level: str = "WARNING",
    min_space_to_bang: Annotated[
        int,
        Option(
            help=dedent(
                """\
                Some documents are '!' delimited instead of ' ' delimited. This
                messes with some vector embedders.  This option attempts to
                infer whether such a case has occured by computing the ratio of
                ' ' to '!'. By default, if '!' absolutely shows up more
                frequently than ' ', then all ' ' will be replaced by '!'.
                min_space_to_bang can be user specified.  If not specified, will
                default to 1, which implies a 1:1 ratio is acceptable (but 1:2,
                1:3, etc is not).
                """
            )
        ),
    ] = 1,
):
    import litellm

    from brag.db import Database
    from brag.emb import LiteLLMEmbeddings
    from brag.repl import BragREPL

    set_logging(log_level)
    if not ssl_verify:
        litellm.ssl_verify = False

    embedder = LiteLLMEmbeddings(
        model=emb,
        api_base=emb_base_url,
        api_key=emb_api_key,
    )

    db = Database(
        corpus_dir=docs_dir,
        embedder=embedder,
        index_doc_batch_size=batchsize,
        num_retrieved_docs=num_retrieved_docs,
        chunk_overlap=chunk_overlap,
        chunk_size=chunk_size,
        cache_db=cache_db,
        db_dir=db_dir,
        db_name=db_name,
        db_host=db_host,
        db_port=db_port,
        terminal=True,
        min_space_to_bang=min_space_to_bang,
    )

    def print_results(query: str):
        print(db.retrieve(query, None))

    BragREPL().run("Begin searching your corpus!", print_results)


@app.command()
def chat(
    llm: Annotated[
        str,
        Option(
            help=dedent(
                """\
                LLM in litellm format (provider/model_id). e.g.,
                * openai/o3
                * ollama/gpt-oss:120b
                * hosted_vllm/meta-llama/Llama-3.1-8B-Instruct
                """
            )
        ),
    ] = "openai/o3",
    llm_base_url: Annotated[
        Optional[str],
        Option(
            help=dedent(
                """\
                Base url for embedder. For example, 
                * https://api.openai.com/v1 (openai)
                * http://localhost:11434/v1 (ollama)
                """
            )
        ),
    ] = None,
    llm_api_key: Annotated[
        Optional[str],
        Option(
            envvar=["BRAG_LLM_API_KEY", "OPENAI_API_KEY"],
            help="API key for embedder",
        ),
    ] = None,
    temperature: Annotated[
        Optional[float], Option(help="Temperature for LLM.", max=1.0, min=0.0)
    ] = None,
    max_completion_tokens: Optional[int] = None,
    max_tokens: Optional[int] = None,
    ssl_verify: Annotated[
        bool, Option(help="Whether or not to enable ssl checks")
    ] = True,
):
    import litellm
    from langchain_litellm import ChatLiteLLM

    from brag.chat import Chatbot
    from brag.repl import BragREPL

    if not ssl_verify:
        litellm.ssl_verify = False

    chat_model = ChatLiteLLM(
        model=llm,
        api_base=llm_base_url,
        api_key=llm_api_key,
        temperature=temperature,
        max_completion_tokens=max_completion_tokens,
        max_tokens=max_completion_tokens,
    )
    chatbot = Chatbot(chat_model)
    BragREPL().run(f"Begin chatting with {llm} ", chatbot.print_stream)


@app.command()
def rm_index(
    db_dir: Annotated[
        Path, Option("--db", help="Directory to store vector database")
    ] = Path(".brag") / "db",
    db_name: Annotated[
        str, Option(help="Database collection name")
    ] = "brag-db",
    db_host: Annotated[
        Optional[str],
        Option(
            help=dedent("""\
                Hostname for chromadb server. If not provided and
                --db-port is provided, this will default to 'localhost'.
                If --db-port is not provided, this flag is ignored. 
                """)
        ),
    ] = None,
    db_port: Annotated[
        Optional[int],
        Option(
            help=dedent("""\
                Port for chromadb server. If not supplied, chromadb will
                run locally.
                """)
        ),
    ] = None,
    reset: Annotated[
        bool,
        Option(
            help="If present, all collections in vector database are removed.",
        ),
    ] = False,
    log_level: str = "WARNING",
):
    from chromadb import HttpClient, PersistentClient
    from chromadb.config import Settings

    set_logging(log_level)

    # Recursively remove db dir.
    local_files_to_remove = [
        db_dir / "corpus_mtimes.sqlite3",
        db_dir / "index-info.txt",
    ]
    for file in local_files_to_remove:
        try:
            file.unlink()
            logging.info(f"Successfully deleted '{file}'.")
        except FileNotFoundError:
            logging.info(f"{file} was not found.")

    # Remove collection if needed.
    settings = Settings(allow_reset=True)
    if db_port is None:
        client = PersistentClient(path=str(db_dir), settings=settings)
    else:
        host = db_host or "localhost"
        client = HttpClient(host=host, port=db_port, settings=settings)

    try:
        # FIXME: the actual directories aren't removed, though the
        # collections are no longer associated. This could lead to large
        # disk space consumption.
        print(
            "Number of collections (before):",
            client.count_collections(),
        )

        client.delete_collection(name=db_name)
        if reset:
            print("HERE")
            client.reset()

        print(
            "Number of collections (after):",
            client.count_collections(),
        )

        logging.info(f"Collection '{db_name}' deleted successfully.")
    except ValueError:
        logging.info(
            f"Collection '{db_name}' was not found at {host}:{db_port}."
        )


@app.command()
def version(short: bool = False):
    if short:
        print(__version__)
    else:
        print(f"brag version {__version__}")


@app.command()
def update_db():
    print("WIP")


def main():
    app()



from rich.console import Console
from rich.theme import Theme


def make_console():
    return Console(theme=Theme({"info": "dim green"}))


chat_css = r"""
:root {
    --font: "Helvetica Neue", Helvetica, Arial, sans-serif;
    --font-weight: 200;
    --bold-weight: 600;
}
:not(.component-wrap).flex-wrap, textarea {
    border-style: none;
}
.chatbot-outer { border-style: none !important; }
.bubble-wrap, .gradio-container {
    background-color: var(--background-fill-primary);
}
div.form, .user-input {
    background-color: var(--background-fill-primary);
    border-style: none !important;
}
strong { font-weight: var(--bold-weight); }
h1, h2, h3, h4, h5, h6 { font-weight: var(--header-weight); }
textarea { resize: none; }
body { padding-bottom: 0em !important; }
.brag-sources li {padding-top: 1em;}
.brag-tab-header {text-align: center;}
.user-input {
    position: fixed;
    bottom: 3em;
    margin-top: 10em;
    left: 50%;
    transform: translateX(-50%);
    max-width: min(85vw, 1000px);
}
"""



import os
from cmd import Cmd
from typing import Callable

from rich import print as rprint

PrintFunction = Callable[[str], None]


class BragREPL(Cmd):
    print_fn: PrintFunction
    prompt: str = "brag> "
    exit_message: str = "Exiting brag shell ..."
    _help_message: str = "For help, type: ? or help"
    ascii = "\n".join([
        r"    __                    ",
        r"   / /_  _________ _____ _",
        r"  / __ \/ ___/ __ `/ __ `/",
        r" / /_/ / /  / /_/ / /_/ / ",
        r"/_.___/_/   \__,_/\__, /  ",
        r"                 /____/   ",
    ])

    def default(self, query: str):
        self.print_fn(query)

    def postcmd(self, stop: bool, line: str):
        print()
        return stop

    def do_exit(self, _):
        """Exit shell."""
        print(self.exit_message)
        return True

    def do_EOF(self, _):
        """Exit on Ctrl+D."""
        print(self.exit_message)
        return True

    def do_clear(self, arg):
        """Clear the screen. Same as pressing Ctrl+L."""
        os.system("cls" if os.name == "nt" else "clear")

    def emptyline(self):
        """Do nothing when an empty line is entered"""
        pass

    def run(self, intro: str, print_fn: PrintFunction):
        """Handle Ctrl+C to avoid quitting the program"""
        # Print intro only once.
        rprint(f"[bold magenta]{self.ascii}")
        print(intro)
        rprint("[grey50]" + self._help_message)
        self.print_fn = print_fn

        while True:
            try:
                self.cmdloop()
                break  # Allows breaking out of loop if EOF is triggered.
            except KeyboardInterrupt:
                print(
                    "\n(Interrupted) Press Ctrl+D to exit or continue typing."
                )


class AskREPL(BragREPL):
    prompt = "brag-ask> "

    def help_refresh(self):
        print("To clear memory, type: !refresh")

    def help_cite(self):
        print("To show sources, type: !cite")
